//krokovy motor
#define EN 8    //enable pin 
#define DIR 5   //smer
#define STEP 2  //krok

//pro displej:
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#define I2C_ADDR  0x27 //adresa I2C prevodniku
#define podsviceni 3
#define En_pin  2
#define Rw_pin  1
#define Rs_pin  0
#define D4_pin  4
#define D5_pin  5
#define D6_pin  6
#define D7_pin  7
LiquidCrystal_I2C  lcd(I2C_ADDR, En_pin, Rw_pin, Rs_pin, D4_pin, D5_pin, D6_pin, D7_pin); //nastaveni displeje
byte stredL[8] = {0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0, 0xf0};
byte stredP[8] = {0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1, 0x1};
byte policko_znak[8] = {0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F};
int rozdil = 100;
float kroky;
int pocetPolicek;
int p; //pocitadlo pri zobrazovani policek

//klavesnice
const int G = 6;
const int C = 7;
const int E = 9; 
const int A = 10;
boolean stiskG = false;
boolean stiskC = false;
boolean stiskE = false;
boolean stiskA = false;
char struna = 0;
int pole;

//preruseni
byte newData = 0;
byte prevData = 0;
unsigned int time = 0;
int timer[10];
int slope[10];
unsigned int totalTimer;
unsigned int perioda;
byte index = 0;
int maxSlope = 0;
int newSlope;
byte noMatch = 0;
byte slopeTol = 3; 
int timerTol = 5;
unsigned int ampTimer = 0;
byte maxAmp = 0;
byte checkMaxAmp;
byte ampThreshold = 5; //promenna pro eliminaci sumu kolem stredni hodnoty, zvoleno experimentalne 

//program
boolean naladeno;
float frekvstup;
double frekvence;
float frekvypocet[30];
int frekvyskyt[30];
byte frekMAX = 0; //cetnost
byte frekINDEX;  //cetnost
byte i;
byte j;
int m;
double frekRef;
float frekvenceG = 392;
float frekvenceC = 261.63;
float frekvenceE = 329.63;
float frekvenceA = 440;
float frekTOL = 2.5; //tolerance přesnosti porovnani frekvence struny a referencni frekvence
int x;
int pozice;
int prePozice=0;
int preRozdil=0;

//PS regulator
float Kp=6;
float Ki=12;
float error;
float suma=0;
float T=0.008;
float output;

void setup() {
  //nastaveni LCD
  lcd.begin (16, 2); //LCD 16x2
  lcd.createChar(1, stredP);
  lcd.createChar(2, stredL);
  lcd.createChar(3, policko_znak);
  lcd.setBacklightPin(podsviceni, POSITIVE); //spusteni podsviceni displeje
  lcd.setBacklight(HIGH);
  lcd.home ();

  //uvitani pri zapnuti
  lcd.print("    VITEJTE!");
  delay(1500);

  //nastaveni vstupu z klavesnice
  pinMode(G, INPUT_PULLUP);
  pinMode(C, INPUT_PULLUP);
  pinMode(E, INPUT_PULLUP);
  pinMode(A, INPUT_PULLUP);

  //nastaveni motoru
  pinMode(EN, OUTPUT);
  pinMode(DIR, OUTPUT);
  pinMode(STEP, OUTPUT);

  //nastaveni A/D prevodniku na portu A3 pro zvyseni vzorkovaci frekvence na 38,5Hz

  cli();//znemozni preruseni

  ADMUX = 0x43; //nastaví A3 jako vstup

  ADCSRA = 0; //vycisti registr ADCSRA

  ADMUX |= (1 << REFS0); //nastavi referenci voltaz 5V
  ADMUX |= (1 << ADLAR); //zarovnani vysledku A/D prevodu doleva

  ADCSRA |= (1 << ADPS2) | (1 << ADPS0); //nastavi prescaler 32
  ADCSRA |= (1 << ADATE); //umozni automaticke spusteni A/D prevodu
  ADCSRA |= (1 << ADIE); //umozni preruseni po A/D prevodu
  ADCSRA |= (1 << ADEN); //umozni A/D prevod
  ADCSRA |= (1 << ADSC); //spusti A/D prevod

  sei();//umozni preruseni
} //setup

//preddefinovane funkce  
void reset() { //cisteni promennych u preruseni
  index = 0;
  noMatch = 0;
  maxSlope = 0;
}

void regulator(){ //PS regulator
  suma=suma+(error);
  output=(Kp*error)+(Ki*T*suma);
}

void cetnost() { //cetnost() pocita nejcastejsi vyskyt frekvence ze vstupu, tim eliminuje sum a vymykajici se hodnoty
  frekINDEX = 0;
  frekMAX = 0;
  for (i = 0; i < 30; i++) {
    frekvyskyt[i] = 0;
  }
  for (i = 0; i < 30; i++) {
    for (j = 0; j < 30; j++) {
      if (frekvypocet[i] == frekvypocet[j]) {
        frekvyskyt[i]++;
      }
    }//for j
  }//for i
  for (i = 0; i < 30; i++) {
    if (frekvyskyt[i] > frekMAX) {
      frekMAX = frekvyskyt[i];
      frekINDEX = i;
    }
  }
  frekvence = frekvypocet[frekINDEX];
}


void policko() {
  lcd.setCursor(p, 2);
  lcd.print((char)3);
  delay(15);
}

void ladeniLCD() {
  lcd.clear();
  lcd.print(struna);
  lcd.setCursor(10, 0);
  lcd.print("LADENI");
  lcd.setCursor(7, 0);
  lcd.print((char)1);
  lcd.setCursor(8, 0);
  lcd.print((char)2);
  lcd.setCursor(7, 1);
  lcd.print((char)1);
  lcd.setCursor(8, 1);
  lcd.print((char)2);
}
//konec preddefinovanych funkci 


//preruseni
ISR(ADC_vect) {// A3 načítá hodnoty každých 26us, v momentě načítání proběne přerušení, která přeruší loop a pak pokračuje cyklus dál
  prevData = newData;//ulozeni predchozi hodnoty
  newData = ADCH;//ziskani hodnoty z A3
  if (prevData < 127 && newData >= 127) { //if prekroceni stredni hodnoty
    newSlope = newData - prevData;//vypocet smernice primky
    if (abs(newSlope - maxSlope) < slopeTol) { //nalezeni nejvyssi smernice primky
      //ulozeni dat do poli 
      slope[index] = newSlope;
      timer[index] = time;
      time = 0;
      if (index == 0) { //pro prvni beh, kdy neni co porovnavat
        noMatch = 0;
        index++;
      }
      else if (abs(timer[0] - timer[index]) < timerTol && abs(slope[0] - newSlope) < slopeTol) { //if casy mezi prechodem stredni hodnoty == a smernice primky ==
        //sum timer values
        totalTimer = 0;
        for (byte i = 0; i < index; i++) {
          totalTimer += timer[i];
        }
        perioda = totalTimer;//urceni periody
        //reset hodnot pro nove hledani periody
        timer[0] = timer[index];
        slope[0] = slope[index];
        index = 1;
        noMatch = 0;
      }
      else { //prekroceni stredni hodnoty bez shody
        index++;
        if (index > 9) { //pokud se po 10 cyklu nepotka s uspechem, dojde k resetu
          reset(); 
        }
      }
    } //vztahuje se k druhemu if
    else if (newSlope > maxSlope) { //nastaveni max. smernice pro nalezeni nejvyssich smernic
      maxSlope = newSlope;
      time = 0;
      noMatch = 0;
      index = 0;
    }

    else { //smernice neni dostatncne vysoka 
      noMatch++;
      if (noMatch > 9) {
        reset();
      }
    }
  } //konec prvniho if

  time++;//pricita hodnotu pri kazdych 26 us
  ampTimer++;
  if (abs(127 - ADCH) > maxAmp) { //nalezeni nejvyssi amplitudy prave vzorkovanych hodnot
                                  //pozdeji slouzi k eliminaci sumu s nizkou amplitudou, ktery by mohl rusit signal      
    maxAmp = abs(127 - ADCH);
  }
  if (ampTimer == 1000) {
    ampTimer = 0;
    checkMaxAmp = maxAmp;
    maxAmp = 0;
  }
}
//konec preruseni 


void loop() {
  digitalWrite(EN, HIGH);
  //vyber struny
  lcd.clear(); //vysicteni displeje
  lcd.print(" Vyberte strunu");
  lcd.setCursor(0,1);
  lcd.print("   k naladeni");
  do {
    if (digitalRead(G) == LOW) {
      stiskG = true; //při stisknutí je LOW
    } else {
      stiskG = false;
    }
    if (digitalRead(C) == LOW) {
      stiskC = true;
    } else {
      stiskC = false;
    }
    if (digitalRead(E) == LOW) {
      stiskE = true;
    } else {
      stiskE = false;
    }
    if (digitalRead(A) == LOW) {
      stiskA = true;
    } else {
      stiskA = false;
    }
  } while (stiskG == false && stiskC == false && stiskE == false && stiskA == false);

//v zavislosti na zvolene strune zavadene ref. frekvence apod.
  if (stiskG == true) {
    lcd.print(struna = 'G');
    frekRef = frekvenceG;
  }
  else if (stiskC == true) {
    lcd.print(struna = 'C');
    frekRef = frekvenceC;
  }
  else if (stiskE == true) {
    lcd.print(struna = 'E');
    frekRef = frekvenceE;
  }
  else if (stiskA == true) {
    lcd.print(struna = 'A');
    frekRef = frekvenceA;
  }

  ladeniLCD(); //obrazovka vizualizace naladeni
  naladeno = false;
  while (naladeno == false) { //cyklus probihajici, dokud neni zase naladeno
    if (checkMaxAmp > ampThreshold) { //pokud je amplituda signalu vyssi nez amplituda sumu
      for (int i = 0; i < 30; i++) { //nacteni 30 hodnot frekvence 
        frekvstup = 38462 / float(perioda); 
        delay(15);
        frekvypocet[i] = frekvstup;
      } //for

      cetnost(); //tato funkce urci nejcentnejsi frekvenci
      rozdil = frekvence - frekRef; 
      pocetPolicek = rozdil / 3;
      if (abs(pocetPolicek) > 7) {
          pocetPolicek = 7;
        }
      
      //graficke zobrazeni a ladeni
      if((abs(rozdil)>=frekTOL)&&(abs(rozdil)<=150)){
        error=abs(rozdil);
        regulator(); //vypocet kroku regulatorem
        kroky=output; 

        if(abs(rozdil)>abs(preRozdil)){prePozice=0;}
        if((rozdil<0&&preRozdil>0)||(rozdil>0&&preRozdil<0)){
          prePozice=0;
          preRozdil=0;
          ladeniLCD();
        }
        //vizualizace a smer otaceni
        if (rozdil < 0) {
          pozice=7-abs(pocetPolicek);
          
          if(prePozice==0){
            for (p = 7; p >= pozice; p--) {
              policko();
            }
          }
          else{
            for(x=prePozice; x<=pozice; x++){
              lcd.setCursor(x, 2);
              lcd.print(" ");
              delay(10);
            }
          }
          digitalWrite(DIR, HIGH);
        }
        else if (rozdil > 0) {
          pozice=8+abs(pocetPolicek);
          if(prePozice==0){
            for (p = 8; p <= pozice; p++) {
              policko();
            }
          }
          else{
            for(x=prePozice; x>pozice; x--){
              lcd.setCursor(x, 2);
              lcd.print(" ");
              delay(10);
            }
          } 
          digitalWrite(DIR, LOW);
        }

        //otaceni      
        digitalWrite(EN, LOW);
        for (m = 0; m < kroky; m++) {
          digitalWrite(STEP, HIGH);
          delay(1);
          digitalWrite(STEP, LOW);
          delay(1);
        }
        digitalWrite(EN, HIGH);
        delay(300);
        
        //ladeniLCD();
        prePozice=pozice;
        preRozdil=rozdil;
      }
    }//if

    if (abs(rozdil) < frekTOL) {naladeno = true;} //if rozdil frekvenci v toleranci
    else {naladeno = false;}
  } //while

  //obrazovka "NALADENO"
  ladeniLCD();
  lcd.setCursor(10, 0);
  lcd.print("   OK ");
  lcd.setCursor(4, 1);
  lcd.print("NALADENO");
  delay(2500);

  //nastaveni hodnot pred dalsim behem loopu
  prePozice=0;
  suma=0; //vypocet regulatoru
  rozdil = 100;
  frekvence = 0;

}//loop


